/*
    CozyCL, a comfortable OpenCL library
    Copyright (C) 2013  Tamas Bolner
	For more information, visit: http://blog.bolner.hu/2013/02/19/cozycl
	
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <sstream>
#include <sys/stat.h>
#include "CozyCL.hpp"

#define COZYCL_MAX_PLATFORMS 10
#define COZYCL_MAX_DEVICES 20

/**
 * Returns all platforms that can be found on the host machine.
 * The index of the Platform in the vector is the "Platform ID".
 * 
 * @returns vector<string>
 **/
vector<string> CozyCL::getPlatforms() {
	cl_uint num_platfoms;
	cl_platform_id platforms[COZYCL_MAX_PLATFORMS];
	cl_int ret;
	char buff[4096];
	string name, version;

	ret = clGetPlatformIDs(COZYCL_MAX_PLATFORMS, platforms, &num_platfoms);
	if (ret != CL_SUCCESS) {
		CozyCL::throwException("clGetPlatformIDs: ", ret);
	}

	vector<string> result(num_platfoms);

	for (unsigned __int32 i = 0; i < num_platfoms; i++) {
		ret = clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME, 4096, buff, NULL);
		if (ret != CL_SUCCESS) {
			CozyCL::throwException("clGetPlatformInfo: ", ret);
		}
		name = string(buff);

		ret = clGetPlatformInfo(platforms[i], CL_PLATFORM_VERSION, 4096, buff, NULL);
		if (ret != CL_SUCCESS) {
			CozyCL::throwException("clGetPlatformInfo: ", ret);
		}
		version = string(buff);

		result[i] = name + " [Version: " + version + "]";
	}


	return result;
}

/**
 * Private
 * Convert error codes to text messages, throws exception.
 *
 * @param char *prefix
 * @param cl_int errorCode
 **/
void CozyCL::throwException(char *prefix, cl_int errorCode) {
	switch (errorCode) {
		case CL_OUT_OF_HOST_MEMORY: throw(new FException(errorCode, "%sThe host is out of memory.", prefix));
		case CL_INVALID_CONTEXT: throw(new FException(errorCode, "%sInvalid context.", prefix));
		case CL_INVALID_PLATFORM: throw(new FException(errorCode, "%sInvalid platform.", prefix));
		case CL_INVALID_PROGRAM: throw(new FException(errorCode, "%sInvalid program.", prefix));
		case CL_INVALID_KERNEL: throw(new FException(errorCode, "%sInvalid kernel.", prefix));
		case CL_INVALID_ARG_INDEX: throw(new FException(errorCode, "%sInvalid argument index.", prefix));
		case CL_INVALID_ARG_VALUE: throw(new FException(errorCode, "%sInvalid argument value.", prefix));
		case CL_INVALID_KERNEL_NAME: throw(new FException(errorCode, "%sInvalid kernel name.", prefix));
		case CL_INVALID_KERNEL_DEFINITION: throw(new FException(errorCode, "%sInvalid kernel definition.", prefix));
		case CL_INVALID_PROGRAM_EXECUTABLE: throw(new FException(errorCode, "%sInvalid program executable.", prefix));
		case CL_INVALID_PROPERTY: throw(new FException(errorCode, "%sInvalid property.", prefix));
		case CL_INVALID_VALUE: throw(new FException(errorCode, "%sInvalid value.", prefix));
		case CL_INVALID_EVENT: throw(new FException(errorCode, "%sInvalid event.", prefix));
		case CL_INVALID_OPERATION: throw(new FException(errorCode, "%sInvalid operation.", prefix));
		case CL_MISALIGNED_SUB_BUFFER_OFFSET: throw(new FException(errorCode, "%sMisaligned sub buffer offset.", prefix));
		case CL_INVALID_EVENT_WAIT_LIST: throw(new FException(errorCode, "%sInvalid event wait list.", prefix));
		case CL_INVALID_MEM_OBJECT: throw(new FException(errorCode, "%sInvalid mem object.", prefix));
		case CL_INVALID_COMMAND_QUEUE: throw(new FException(errorCode, "%sInvalid command queue.", prefix));
		case CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST: throw(new FException(errorCode, "%sThe execution status is error for one or more of the events in the wait list.", prefix));
		case CL_INVALID_DEVICE_TYPE: throw(new FException(errorCode, "%sInvalid device type.", prefix));
		case CL_INVALID_DEVICE: throw(new FException(errorCode, "%sInvalid device.", prefix));
		case CL_MEM_OBJECT_ALLOCATION_FAILURE: throw(new FException(errorCode, "%sFailure while allocating memory for buffer.", prefix));
		case CL_INVALID_HOST_PTR: throw(new FException(errorCode, "%sInvalid host pointer.", prefix));
		case CL_INVALID_BUFFER_SIZE: throw(new FException(errorCode, "%sInvalid buffer size.", prefix));
		case CL_INVALID_QUEUE_PROPERTIES: throw(new FException(errorCode, "%sInvalid queue properties.", prefix));
		case CL_DEVICE_NOT_FOUND: throw(new FException(errorCode, "%sDevice not found.", prefix));
		case CL_DEVICE_NOT_AVAILABLE: throw(new FException(errorCode, "%sDevice not available.", prefix));
		case CL_OUT_OF_RESOURCES: throw(new FException(errorCode, "%sOut of resources.", prefix));
	}

	throw(new FException("%sError code: %d", prefix, errorCode));
}

/**
 * Returns the Devices for the given Platform.
 * The index of a device inside the returned array is its "Device ID" (inside the Platform)
 * 
 * @param uInt32 platform_id See -> CozyCL::getPlatforms()
 * @returns vector<string>
 **/
vector<string> CozyCL::getDevices(uInt32 platform_id) {
	cl_uint num_platfoms;
	cl_platform_id platforms[COZYCL_MAX_PLATFORMS];
	cl_int ret;
	char buff[4096];
	string name, vendor;
	cl_uint vendor_id;
	cl_device_id devices[COZYCL_MAX_DEVICES];
	cl_uint num_devices;
	ostringstream Convert;
	uInt32 i;

	ret = clGetPlatformIDs(COZYCL_MAX_PLATFORMS, platforms, &num_platfoms);
	if (ret != CL_SUCCESS) {
		CozyCL::throwException("clGetPlatformIDs: ", ret);
	}

	if (platform_id >= num_platfoms) {
		throw(new FException("CozyCL::getDevices: Invalid platform index: %u", platform_id));
	}

	ret = clGetDeviceIDs(platforms[platform_id], CL_DEVICE_TYPE_ALL, COZYCL_MAX_DEVICES, devices, &num_devices);
	if (ret != CL_SUCCESS) {
		CozyCL::throwException("clGetDeviceIDs: ", ret);
	}

	vector<string> result(num_devices);

	for (i = 0; i < num_devices; i++) {
		ret = clGetDeviceInfo(devices[i], CL_DEVICE_NAME, 4096, buff, NULL);
		if (ret != CL_SUCCESS) {
			CozyCL::throwException("clGetDeviceInfo: ", ret);
		}
		name = string(buff);

		ret = clGetDeviceInfo(devices[i], CL_DEVICE_VENDOR, 4096, buff, NULL);
		if (ret != CL_SUCCESS) {
			CozyCL::throwException("clGetDeviceInfo: ", ret);
		}
		vendor = string(buff);

		ret = clGetDeviceInfo(devices[i], CL_DEVICE_VENDOR_ID, 4, &vendor_id, NULL);
		if (ret != CL_SUCCESS) {
			CozyCL::throwException("clGetDeviceInfo: ", ret);
		}
		Convert.clear();
		Convert << vendor_id;

		result[i] = name + " - " + vendor + " (Vendor ID: " + Convert.str() + ")";
	}

	return result;
}

/**
 * Constructor.
 * Creates OpenCL objects.
 *
 * @param uInt32 platform_id
 * @param uInt32 device_id
 * @param char *program_FilePath
 * @param char *kernel_func
 * @param uInt32 memoryObject_num
 * @param uInt64 memoryObject_memSize
 * @returns CozyCL
 **/
CozyCL::CozyCL(uInt32 platform_id, uInt32 device_id, char *program_FilePath, char *kernel_func, uInt32 memoryObject_num, uInt64 memoryObject_memSize) {
	cl_uint num_platfoms;
	cl_platform_id platforms[COZYCL_MAX_PLATFORMS];
	cl_int ret;
	cl_device_id devices[COZYCL_MAX_DEVICES];
	cl_uint num_devices;
	cl_mem buffer;
	cl_kernel kernel;
	cl_event fill_finished_event;
	unsigned char fill_pattern = 0;

	this->cz_platform_id = NULL;
	this->cz_device_id = NULL;
	this->cz_context = NULL;
	this->cz_queue = NULL;
	this->cz_program = NULL;
	this->cz_buffers = NULL;
	this->cz_kernels = NULL;
	this->cz_events = NULL;

	if (memoryObject_num < 1) {
		throw(new FException("CozyCL: The memory object number must be larger than zero."));
	}
	this->cz_memoryObject_num = memoryObject_num;

	try {
		/*
			Platform
		*/
		ret = clGetPlatformIDs(COZYCL_MAX_PLATFORMS, platforms, &num_platfoms);
		if (ret != CL_SUCCESS) {
			CozyCL::throwException("clGetPlatformIDs: ", ret);
		}

		if (platform_id >= num_platfoms) {
			throw(new FException("CozyCL: Invalid platform id: %u", platform_id));
		}

		this->cz_platform_id = platforms[platform_id];

		/*
			Device
		*/
		ret = clGetDeviceIDs(platforms[platform_id], CL_DEVICE_TYPE_ALL, COZYCL_MAX_DEVICES, devices, &num_devices);
		if (ret != CL_SUCCESS) {
			CozyCL::throwException("clGetDeviceIDs: ", ret);
		}

		if (device_id >= num_devices) {
			throw(new FException("CozyCL: Invalid device id: %d", device_id));
		}

		this->cz_device_id = devices[device_id];

		/*
			Context
		*/
		this->cz_context = clCreateContext(NULL, 1, &this->cz_device_id, NULL, NULL, &ret);
		if (ret != CL_SUCCESS) {
			CozyCL::throwException("clCreateContext: ", ret);
		}

		/*
			Command queue (out-of-order)
		*/
		this->cz_queue = clCreateCommandQueue(this->cz_context, this->cz_device_id, CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE, &ret);
		if (ret != CL_SUCCESS) {
			CozyCL::throwException("clCreateCommandQueue: ", ret);
		}

		/*
			Program
		*/
		this->cz_program = this->build_program(this->cz_context, this->cz_device_id, program_FilePath);

		/*
			Create objects per memory object:
				- buffer
				- fill buffer with zeroes
				- kernel
				- set argument
		*/
		this->cz_buffers = new cl_mem[memoryObject_num];
		for (uInt32 i = 0; i < memoryObject_num; i++) this->cz_buffers[i] = NULL;

		this->cz_kernels = new cl_kernel[memoryObject_num];
		for (uInt32 i = 0; i < memoryObject_num; i++) this->cz_kernels[i] = NULL;

		this->cz_events = new cl_event[memoryObject_num];		// this'll get filled in the startKernel() method
		for (uInt32 i = 0; i < memoryObject_num; i++) this->cz_events[i] = NULL;

		for (uInt32 i = 0; i < memoryObject_num; i++) {
			//Create Buffer
			buffer = clCreateBuffer(this->cz_context, CL_MEM_READ_WRITE, memoryObject_memSize, NULL, &ret);
			if (ret != CL_SUCCESS) {
				CozyCL::throwException("clCreateBuffer: ", ret);
			}
			this->cz_buffers[i] = buffer;

			//Fill buffer with zeroes
			ret = clEnqueueFillBuffer(this->cz_queue, buffer, &fill_pattern, sizeof(fill_pattern), 0, memoryObject_memSize, 0, NULL, &fill_finished_event);
			if (ret != CL_SUCCESS) {
				CozyCL::throwException("clEnqueueFillBuffer: ", ret);
			}

			//Wait for the filling to complete (the queue is in out-of-order mode currently)
			ret = clWaitForEvents(1, &fill_finished_event);
			if (ret != CL_SUCCESS) {
				CozyCL::throwException("clWaitForEvents (fill): ", ret);
			}

			//Create Kernel
			kernel = clCreateKernel(this->cz_program, kernel_func, &ret);
			if (ret != CL_SUCCESS) {
				CozyCL::throwException("clCreateKernel: ", ret);
			}
			this->cz_kernels[i] = kernel;

			//Set kernel arguments
			ret = clSetKernelArg(kernel, 0, sizeof(cl_mem), &buffer);
			if (ret != CL_SUCCESS) {
				CozyCL::throwException("clSetKernelArg: ", ret);
			}

			ret = clSetKernelArg(kernel, 1, sizeof(memoryObject_memSize), &memoryObject_memSize);
			if (ret != CL_SUCCESS) {
				CozyCL::throwException("clSetKernelArg: ", ret);
			}
		}
	} catch (FException *ex) {
		/*
			Cleanup and re-throw.
		*/
		this->cleanup();
		throw(ex);
	}
}

/**
 * Private
 * Builds the program.
 *
 * @param cl_context context
 * @param cl_device_id device_id
 * @param *file_path
 * @returns cl_program
 **/
cl_program CozyCL::build_program(cl_context context, cl_device_id device_id, char *file_path) {
	struct stat filestatus;
	FILE *stream;
	errno_t result;
	char *program_txt, *program_log;
	size_t bytes_read, log_size;
	cl_int ret;
	cl_program program;
	FException *ex;

	/*
		Load file
	*/
	if (stat(file_path, &filestatus) != 0) {
		throw(new FException("OpenCL program file not found: '%s'", file_path));
	}

	result = fopen_s(&stream, file_path, "r");
	if (result != 0) {
		throw(new FException("Can't open program file: '%s'", file_path));
	}

	program_txt = (char *)malloc(filestatus.st_size + 1);
	if (program_txt == NULL) {
		fclose(stream);
		throw(new FException("Can't allocate memory for '%s'. Required amount: %ld bytes.", file_path, filestatus.st_size + 1));
	}

	bytes_read = fread(program_txt, 1, filestatus.st_size, stream);
	fclose(stream);

	if (bytes_read < 1) {
		throw(new FException("Can't read file: '%s'", file_path));
	}

	*(program_txt + bytes_read) = 0;

	/*
		Create program
	*/
	program = clCreateProgramWithSource(context, 1, (const char**)&program_txt, &bytes_read, &ret);
	if (ret != CL_SUCCESS) {
		free(program_txt);
		CozyCL::throwException("clCreateProgramWithSource: ", ret);
	}
	free(program_txt);

	/*
		Build program
	*/
	ret = clBuildProgram(program, 0, NULL, NULL, NULL, NULL);
	if (ret != CL_SUCCESS) {
		/*
			Get build log
		*/
		clGetProgramBuildInfo(program, device_id, CL_PROGRAM_BUILD_LOG, 0, NULL, &log_size);
		program_log = (char*) malloc(log_size + 1);
		clGetProgramBuildInfo(program, device_id, CL_PROGRAM_BUILD_LOG, log_size, program_log, NULL);
		program_log[log_size] = 0;
		ex = new FException("Program build failed. File: '%s'. Log: \n\n%s", file_path, program_log);

		clReleaseProgram(program);
		free(program_log);

		throw(ex);
	}

	return program;
}

/**
 * Private
 * Deallocates all OpenCL objects.
 **/
void CozyCL::cleanup() {
	if (this->cz_events != NULL) {
		this->waitForAll();

		delete[] this->cz_events;
		this->cz_events = NULL;
	}

	if (this->cz_kernels != 0) {
		for (uInt32 i = 0; i < this->cz_memoryObject_num; i++) {
			if (this->cz_kernels[i] != NULL) {
				clReleaseKernel(this->cz_kernels[i]);
				this->cz_kernels[i] = NULL;
			}
		}

		delete[] this->cz_kernels;
		this->cz_kernels = NULL;
	}
	
	if (this->cz_buffers != 0) {
		for (uInt32 i = 0; i < this->cz_memoryObject_num; i++) {
			if (this->cz_buffers[i] != NULL) {
				clReleaseMemObject(this->cz_buffers[i]);
				this->cz_buffers[i] = NULL;
			}
		}

		delete[] this->cz_buffers;
		this->cz_buffers = NULL;
	}
	
	if (this->cz_queue != NULL) {
		clReleaseCommandQueue(this->cz_queue);
		this->cz_queue = NULL;
	}

	if (this->cz_program != NULL) {
		clReleaseProgram(this->cz_program);
		this->cz_program = NULL;
	}

	if (this->cz_context != NULL) {
		clReleaseContext(this->cz_context);
		this->cz_context = NULL;
	}
}

/**
 * Destructor
 * Deallocates all OpenCL objects.
 **/
CozyCL::~CozyCL() {
	this->cleanup();
}

/**
 * Copies a part of the host memory into the memory of the device.
 *
 * @param uInt32 memoryObject_id
 * @param void *host_addr
 * @param uInt64 size
 * @param uInt64 device_addr
 **/
void CozyCL::writeMemory(uInt32 memoryObject_id, void *host_addr, uInt64 size, uInt64 device_addr) {
	cl_uint ret;

	if (memoryObject_id >= this->cz_memoryObject_num) {
		throw(new FException("CozyCL::writeMemory: Invalid memory object id: %u", memoryObject_id));
	}

	//Blocking write
	ret = clEnqueueWriteBuffer(this->cz_queue, this->cz_buffers[memoryObject_id], CL_TRUE, device_addr, size, host_addr, 0, NULL, NULL);
	if (ret != CL_SUCCESS) {
		CozyCL::throwException("clEnqueueWriteBuffer: ", ret);
	}
}

/**
 * Copies a part of the device memory into the memory of the host machine.
 * 
 * @param uInt32 memoryObject_id
 * @param void *host_addr
 * @param uInt64 size
 * @param uInt64 device_addr
 **/
void CozyCL::readMemory(uInt32 memoryObject_id, void *host_addr, uInt64 size, uInt64 device_addr) {
	cl_uint ret;

	if (memoryObject_id >= this->cz_memoryObject_num) {
		throw(new FException("CozyCL::readMemory: Invalid memory object id: %u", memoryObject_id));
	}

	//Blocking read
	ret = clEnqueueReadBuffer(this->cz_queue, this->cz_buffers[memoryObject_id], CL_TRUE, device_addr, size, host_addr, 0, NULL, NULL);
	if (ret != CL_SUCCESS) {
		CozyCL::throwException("clEnqueueReadBuffer: ", ret);
	}
}

/**
 * Waits until every work item, started by the startKernel() 
 * method gets finished.
 * 
 **/
void CozyCL::waitForAll() {
	cl_uint ret;

	if (this->cz_events != NULL) {
		for (uInt32 i = 0; i < this->cz_memoryObject_num; i++) {
			if (this->cz_events[i] != NULL) {
				ret = clWaitForEvents(1, &this->cz_events[i]);
				if (ret != CL_SUCCESS) {
					CozyCL::throwException("clWaitForEvents (waitForAll): ", ret);
				}
			}
		}
	}
}

/**
 * Deploys the kernel for the device, and creates work items
 * which will work on the selected memory object.
 *
 * @param uInt32 memoryObject_id
 * @param uInt32 global_size
 * @param uInt32 local_size
 **/
void CozyCL::startKernel(uInt32 memoryObject_id, uInt32 global_size, uInt32 local_size) {
	cl_uint ret;
	size_t gs, ls;

	gs = global_size;
	ls = local_size;

	if (memoryObject_id >= this->cz_memoryObject_num) {
		throw(new FException("CozyCL::startKernel: Invalid memory object id: %u", memoryObject_id));
	}

	ret = clEnqueueNDRangeKernel(this->cz_queue, this->cz_kernels[memoryObject_id], 1, NULL, &gs, &ls, 0, NULL, &(this->cz_events[memoryObject_id]));
	if (ret != CL_SUCCESS) {
		CozyCL::throwException("CozyCL::startKernel: ", ret);
	}
}
